@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2011  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -1228,6 +1228,9 @@ static struct connman_iptables *iptables_init(char *table_name)
 	char *module = NULL;
 	socklen_t s;
 
+	if (table_name == NULL)
+		table_name = "filter";
+
 	DBG("%s", table_name);
 
 	if (xtables_insmod("ip_tables", NULL, TRUE) != 0)
@@ -1506,6 +1509,15 @@ out:
 	return err;
 }
 
+static struct connman_iptables *pre_load_table(char *table_name,
+					struct connman_iptables *table)
+{
+	if (table != NULL)
+		return table;
+
+	return iptables_init(table_name);
+}
+
 static int iptables_command(int argc, char *argv[])
 {
 	struct connman_iptables *table;
@@ -1526,14 +1538,15 @@ static int iptables_command(int argc, char *argv[])
 	insert = FALSE;
 	delete = FALSE;
 	compare = FALSE;
-	table_name = chain = new_chain = match_name = target_name = NULL;
-	flush_chain = delete_chain = policy = NULL;
+	chain = new_chain = match_name = target_name = NULL;
+	flush_chain = delete_chain = policy = table_name = NULL;
 	memset(&ip, 0, sizeof(struct ipt_ip));
 	table = NULL;
 	xt_rm = NULL;
 	xt_m = NULL;
 	xt_t = NULL;
-	ret = 0;
+	/* Default code for options parsing */
+	ret = -EINVAL;
 
 	/* extension's options will generate false-positives errors */
 	opterr = 0;
@@ -1629,6 +1642,11 @@ static int iptables_command(int argc, char *argv[])
 
 		case 'j':
 			target_name = optarg;
+
+			table = pre_load_table(table_name, table);
+			if (table == NULL)
+				goto out;
+
 			xt_t = prepare_target(table, target_name);
 			if (xt_t == NULL)
 				goto out;
@@ -1637,6 +1655,11 @@ static int iptables_command(int argc, char *argv[])
 
 		case 'm':
 			match_name = optarg;
+
+			table = pre_load_table(table_name, table);
+			if (table == NULL)
+				goto out;
+
 			xt_m = prepare_matches(table, &xt_rm, match_name);
 			if (xt_m == NULL)
 				goto out;
@@ -1669,11 +1692,9 @@ static int iptables_command(int argc, char *argv[])
 		case 't':
 			table_name = optarg;
 
-			table = iptables_init(table_name);
-			if (table == NULL) {
-				ret = -EINVAL;
+			table = pre_load_table(table_name, table);
+			if (table == NULL)
 				goto out;
-			}
 
 			break;
 
@@ -1686,7 +1707,6 @@ static int iptables_command(int argc, char *argv[])
 
 			connman_error("Invalid option");
 
-			ret = -EINVAL;
 			goto out;
 
 		default:
@@ -1765,15 +1785,12 @@ static int iptables_command(int argc, char *argv[])
 		xt_t->final_check(xt_t->tflags);
 #endif
 
-	if (table == NULL) {
-		table_name = "filter";
+	table = pre_load_table(table_name, table);
+	if (table == NULL)
+		goto out;
 
-		table = iptables_init(table_name);
-		if (table == NULL) {
-			ret = -EINVAL;
-			goto out;
-		}
-	}
+	/* Option parsing went fine, falling back to succes code */
+	ret = 0;
 
 	if (delete_chain != NULL) {
 		printf("Delete chain %s\n", delete_chain);
@@ -1786,7 +1803,6 @@ static int iptables_command(int argc, char *argv[])
 	if (dump) {
 		iptables_dump(table);
 
-		ret = 0;
 		goto out;
 	}
 
