@@ -28,6 +28,19 @@
 
 #include "connman.h"
 
+/*
+ * How many times to send RS with the purpose of
+ * refreshing RDNSS entries before they actually expire.
+ * With a value of 1, one RS will be sent, with no retries.
+ */
+#define RS_REFRESH_COUNT	1
+
+/*
+ * Value in seconds to wait for RA after RS was sent.
+ * After this time elapsed, we can send another RS.
+ */
+#define RS_REFRESH_TIMEOUT	3
+
 static GSList *network_list = NULL;
 static GSList *driver_list = NULL;
 
@@ -46,6 +59,7 @@ struct connman_network {
 	char *path;
 	int index;
 	int router_solicit_count;
+	int router_solicit_refresh_count;
 
 	struct connman_network_driver *driver;
 	void *driver_data;
@@ -450,36 +464,19 @@ void connman_network_set_index(struct connman_network *network, int index)
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto done;
 
 	ipconfig = __connman_service_get_ip4config(service);
+	if (ipconfig == NULL)
+		goto done;
+
+	/* If index changed, the index of ipconfig must be reset. */
+	__connman_ipconfig_set_index(ipconfig, index);
 
 	DBG("index %d service %p ip4config %p", network->index,
 		service, ipconfig);
-
-	if (network->index < 0 && ipconfig == NULL) {
-
-		ipconfig = __connman_service_get_ip4config(service);
-		if (ipconfig == NULL)
-			/*
-			 * This is needed for plugins that havent set their
-			 * ipconfig layer yet, due to not being able to get
-			 * a network index prior to creating a service.
-			 */
-			__connman_service_create_ip4config(service, index);
-		else
-			__connman_ipconfig_set_index(ipconfig, index);
-
-	} else {
-		/* If index changed, the index of ipconfig must be reset. */
-		if (ipconfig == NULL)
-			goto done;
-
-		__connman_ipconfig_set_index(ipconfig, index);
-	}
-
 done:
 	network->index = index;
 }
@@ -633,7 +630,7 @@ int connman_network_set_associating(struct connman_network *network,
 	if (associating == TRUE) {
 		struct connman_service *service;
 
-		service = __connman_service_lookup_from_network(network);
+		service = connman_service_lookup_from_network(network);
 		__connman_service_ipconfig_indicate_state(service,
 					CONNMAN_SERVICE_STATE_ASSOCIATION,
 					CONNMAN_IPCONFIG_TYPE_IPV4);
@@ -649,7 +646,7 @@ static void set_associate_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -659,7 +656,7 @@ static void set_configure_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -669,7 +666,7 @@ static void set_invalid_key_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_INVALID_KEY);
@@ -679,7 +676,7 @@ static void set_connect_error(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_CONNECT_FAILED);
@@ -691,7 +688,7 @@ void connman_network_set_ipv4_method(struct connman_network *network,
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -708,7 +705,7 @@ void connman_network_set_ipv6_method(struct connman_network *network,
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -759,7 +756,7 @@ void connman_network_clear_error(struct connman_network *network)
 	if (network->connecting == TRUE || network->associating == TRUE)
 		return;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	__connman_service_clear_error(service);
 }
 
@@ -777,7 +774,7 @@ static void set_configuration(struct connman_network *network,
 
 	connman_device_set_disconnected(network->device, FALSE);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	__connman_service_ipconfig_indicate_state(service,
 					CONNMAN_SERVICE_STATE_CONFIGURATION,
 					type);
@@ -789,7 +786,7 @@ static void dhcp_success(struct connman_network *network)
 	struct connman_ipconfig *ipconfig_ipv4;
 	int err;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto err;
 
@@ -817,7 +814,7 @@ static void dhcp_failure(struct connman_network *network)
 {
 	struct connman_service *service;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -845,7 +842,7 @@ static int set_connected_fixed(struct connman_network *network)
 
 	DBG("");
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 
@@ -880,7 +877,7 @@ static void set_connected_manual(struct connman_network *network)
 
 	DBG("network %p", network);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig = __connman_service_get_ip4config(service);
 
@@ -934,13 +931,15 @@ static int manual_ipv6_set(struct connman_network *network,
 
 	DBG("network %p ipv6 %p", network, ipconfig_ipv6);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
 	if (__connman_ipconfig_get_local(ipconfig_ipv6) == NULL)
 		__connman_service_read_ip6config(service);
 
+	__connman_ipconfig_enable_ipv6(ipconfig_ipv6);
+
 	err = __connman_ipconfig_address_add(ipconfig_ipv6);
 	if (err < 0) {
 		connman_network_set_error(network,
@@ -997,7 +996,7 @@ static gboolean dhcpv6_set_addresses(struct connman_network *network)
 	struct connman_ipconfig *ipconfig_ipv6;
 	int err = -EINVAL;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		goto err;
 
@@ -1113,6 +1112,58 @@ static void check_dhcpv6(struct nd_router_advert *reply,
 	connman_network_unref(network);
 }
 
+static void receive_refresh_rs_reply(struct nd_router_advert *reply,
+		unsigned int length, void *user_data)
+{
+	struct connman_network *network = user_data;
+
+	DBG("reply %p", reply);
+
+	if (reply == NULL) {
+		/*
+		 * Router solicitation message seem to get lost easily so
+		 * try to send it again.
+		 */
+		if (network->router_solicit_refresh_count > 1) {
+			network->router_solicit_refresh_count--;
+			DBG("re-send router solicitation %d",
+					network->router_solicit_refresh_count);
+			__connman_inet_ipv6_send_rs(network->index,
+					RS_REFRESH_TIMEOUT,
+					receive_refresh_rs_reply,
+					network);
+			return;
+		}
+	}
+
+	/* RS refresh not in progress anymore */
+	network->router_solicit_refresh_count = 0;
+
+	connman_network_unref(network);
+	return;
+}
+
+int __connman_refresh_rs_ipv6(struct connman_network *network, int index)
+{
+	int ret = 0;
+
+	DBG("network %p index %d", network, index);
+
+	/* Send only one RS for all RDNSS entries which are about to expire */
+	if (network->router_solicit_refresh_count > 0) {
+		DBG("RS refresh already started");
+		return 0;
+	}
+
+	network->router_solicit_refresh_count = RS_REFRESH_COUNT;
+
+	connman_network_ref(network);
+
+	ret = __connman_inet_ipv6_send_rs(index, RS_REFRESH_TIMEOUT,
+			receive_refresh_rs_reply, network);
+	return ret;
+}
+
 static void autoconf_ipv6_set(struct connman_network *network)
 {
 	struct connman_service *service;
@@ -1137,7 +1188,7 @@ static void autoconf_ipv6_set(struct connman_network *network)
 
 	network->connecting = FALSE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return;
 
@@ -1166,7 +1217,7 @@ static void set_connected(struct connman_network *network)
 
 	network->connected = TRUE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 	ipconfig_ipv6 = __connman_service_get_ip6config(service);
@@ -1238,7 +1289,7 @@ static void set_disconnected(struct connman_network *network)
 
 	network->connected = FALSE;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 
 	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 	ipconfig_ipv6 = __connman_service_get_ip6config(service);
@@ -1389,17 +1440,36 @@ connman_bool_t connman_network_get_associating(struct connman_network *network)
 	return network->associating;
 }
 
+void connman_network_clear_hidden(void *user_data)
+{
+	if (user_data == NULL)
+		return;
+
+	DBG("user_data %p", user_data);
+
+	/*
+	 * Hidden service does not have a connect timeout so
+	 * we do not need to remove it. We can just return
+	 * error to the caller telling that we could not find
+	 * any network that we could connect to.
+	 */
+	__connman_service_reply_dbus_pending(user_data, EIO, NULL);
+}
+
 int connman_network_connect_hidden(struct connman_network *network,
-				char *identity, char* passphrase)
+			char *identity, char* passphrase, void *user_data)
 {
 	int err = 0;
 	struct connman_service *service;
 
-	DBG("");
+	service = connman_service_lookup_from_network(network);
 
-	service = __connman_service_lookup_from_network(network);
-	if (service == NULL)
-		return -EINVAL;
+	DBG("network %p service %p user_data %p", network, service, user_data);
+
+	if (service == NULL) {
+		err = -EINVAL;
+		goto out;
+	}
 
 	if (identity != NULL)
 		__connman_service_set_agent_identity(service, identity);
@@ -1410,9 +1480,17 @@ int connman_network_connect_hidden(struct connman_network *network,
 	if (err == -ENOKEY) {
 		__connman_service_indicate_error(service,
 					CONNMAN_SERVICE_ERROR_INVALID_KEY);
-		return err;
-	} else
+		goto out;
+	} else {
+		__connman_service_set_hidden(service);
+		__connman_service_set_userconnect(service, TRUE);
+		__connman_service_set_hidden_data(service, user_data);
 		return __connman_service_connect(service);
+	}
+
+out:
+	__connman_service_return_error(service, -err, user_data);
+	return err;
 }
 
 /**
@@ -1499,7 +1577,7 @@ static int manual_ipv4_set(struct connman_network *network,
 	struct connman_service *service;
 	int err;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1520,7 +1598,7 @@ int __connman_network_clear_ipconfig(struct connman_network *network,
 	enum connman_ipconfig_method method;
 	enum connman_ipconfig_type type;
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1616,7 +1694,7 @@ int connman_network_set_ipaddress(struct connman_network *network,
 
 	DBG("network %p", network);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1642,7 +1720,7 @@ int connman_network_set_nameservers(struct connman_network *network,
 
 	DBG("network %p nameservers %s", network, nameservers);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
@@ -1670,7 +1748,7 @@ int connman_network_set_domain(struct connman_network *network,
 
 	DBG("network %p domain %s", network, domain);
 
-	service = __connman_service_lookup_from_network(network);
+	service = connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
