@@ -69,6 +69,46 @@ struct ntp_msg {
 
 #define LOGTOD(a)  ((a) < 0 ? 1. / (1L << -(a)) : 1L << (int)(a))
 
+#define NTP_SEND_TIMEOUT       2
+#define NTP_SEND_RETRIES       3
+
+#define NTP_FLAG_LI_SHIFT      6
+#define NTP_FLAG_LI_MASK       0x3
+#define NTP_FLAG_LI_NOWARNING  0x0
+#define NTP_FLAG_LI_ADDSECOND  0x1
+#define NTP_FLAG_LI_DELSECOND  0x2
+#define NTP_FLAG_LI_NOTINSYNC  0x3
+
+#define NTP_FLAG_VN_SHIFT      3
+#define NTP_FLAG_VN_MASK       0x7
+
+#define NTP_FLAG_MD_SHIFT      0
+#define NTP_FLAG_MD_MASK       0x7
+#define NTP_FLAG_MD_UNSPEC     0
+#define NTP_FLAG_MD_ACTIVE     1
+#define NTP_FLAG_MD_PASSIVE    2
+#define NTP_FLAG_MD_CLIENT     3
+#define NTP_FLAG_MD_SERVER     4
+#define NTP_FLAG_MD_BROADCAST  5
+#define NTP_FLAG_MD_CONTROL    6
+#define NTP_FLAG_MD_PRIVATE    7
+
+#define NTP_FLAGS_ENCODE(li, vn, md)  ((uint8_t)( \
+                      (((li) & NTP_FLAG_LI_MASK) << NTP_FLAG_LI_SHIFT) | \
+                      (((vn) & NTP_FLAG_VN_MASK) << NTP_FLAG_VN_SHIFT) | \
+                      (((md) & NTP_FLAG_MD_MASK) << NTP_FLAG_MD_SHIFT)))
+
+#define NTP_FLAGS_LI_DECODE(flags)    ((uint8_t)(((flags) >> NTP_FLAG_LI_SHIFT) & NTP_FLAG_LI_MASK))
+#define NTP_FLAGS_VN_DECODE(flags)    ((uint8_t)(((flags) >> NTP_FLAG_VN_SHIFT) & NTP_FLAG_VN_MASK))
+#define NTP_FLAGS_MD_DECODE(flags)    ((uint8_t)(((flags) >> NTP_FLAG_MD_SHIFT) & NTP_FLAG_MD_MASK))
+
+#define NTP_PRECISION_S    0
+#define NTP_PRECISION_DS   -3
+#define NTP_PRECISION_CS   -6
+#define NTP_PRECISION_MS   -9
+#define NTP_PRECISION_US   -19
+#define NTP_PRECISION_NS   -29
+
 static guint channel_watch = 0;
 static struct timeval transmit_timeval;
 static int transmit_fd = 0;
@@ -76,6 +116,31 @@ static int transmit_fd = 0;
 static char *timeserver = NULL;
 static gint poll_id = 0;
 static gint timeout_id = 0;
+static guint retries = 0;
+
+static void send_packet(int fd, const char *server);
+
+static void next_server(void)
+{
+	if (timeserver != NULL) {
+		g_free(timeserver);
+		timeserver = NULL;
+	}
+
+	__connman_timeserver_sync_next();
+}
+
+static gboolean send_timeout(gpointer user_data)
+{
+	DBG("send timeout (retries %d)", retries);
+
+	if (retries++ == NTP_SEND_RETRIES)
+		next_server();
+	else
+		send_packet(transmit_fd, timeserver);
+
+	return FALSE;
+}
 
 static void send_packet(int fd, const char *server)
 {
@@ -83,12 +148,17 @@ static void send_packet(int fd, const char *server)
 	struct sockaddr_in addr;
 	ssize_t len;
 
+	/*
+	 * At some point, we could specify the actual system precision with:
+	 *
+	 *   clock_getres(CLOCK_REALTIME, &ts);
+	 *   msg.precision = (int)log2(ts.tv_sec + (ts.tv_nsec * 1.0e-9));
+	 */
 	memset(&msg, 0, sizeof(msg));
-	msg.flags = 0x23;
+	msg.flags = NTP_FLAGS_ENCODE(NTP_FLAG_LI_NOTINSYNC, 4, NTP_FLAG_MD_CLIENT);
 	msg.poll = 4;	// min
 	msg.poll = 10;	// max
-	msg.xmttime.seconds = random();
-	msg.xmttime.fraction = random();
+	msg.precision = NTP_PRECISION_S;
 
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
@@ -97,10 +167,18 @@ static void send_packet(int fd, const char *server)
 
 	gettimeofday(&transmit_timeval, NULL);
 
+	msg.xmttime.seconds = htonl(transmit_timeval.tv_sec + OFFSET_1900_1970);
+	msg.xmttime.fraction = htonl(transmit_timeval.tv_usec * 1000);
+
 	len = sendto(fd, &msg, sizeof(msg), MSG_DONTWAIT,
 						&addr, sizeof(addr));
 	if (len < 0) {
-		connman_error("Time request for server %s failed", server);
+		connman_error("Time request for server %s failed (%d/%s)",
+			server, errno, strerror(errno));
+
+		if (errno == ENETUNREACH)
+			__connman_timeserver_sync_next();
+
 		return;
 	}
 
@@ -108,18 +186,14 @@ static void send_packet(int fd, const char *server)
 		connman_error("Broken time request for server %s", server);
 		return;
 	}
-}
 
-static gboolean next_server(gpointer user_data)
-{
-	if (timeserver != NULL) {
-		g_free(timeserver);
-		timeserver = NULL;
-	}
+	/*
+	 * Add a retry timeout of two seconds to retry the existing
+	 * request. After a set number of retries, we'll fallback to
+	 * trying another server.
+	 */
 
-	__connman_timeserver_sync_next();
-
-	return FALSE;
+	timeout_id = g_timeout_add_seconds(NTP_SEND_TIMEOUT, send_timeout, NULL);
 }
 
 static gboolean next_poll(gpointer user_data)
@@ -132,6 +206,14 @@ static gboolean next_poll(gpointer user_data)
 	return FALSE;
 }
 
+static void reset_timeout(void)
+{
+	if (timeout_id > 0)
+		g_source_remove(timeout_id);
+
+	retries = 0;
+}
+
 static void decode_msg(void *base, size_t len, struct timeval *tv)
 {
 	struct ntp_msg *msg = base;
@@ -163,8 +245,20 @@ static void decode_msg(void *base, size_t len, struct timeval *tv)
 
 	transmit_delay = LOGTOD(msg->poll);
 
-	if (msg->flags != 0x24)
+	if (NTP_FLAGS_LI_DECODE(msg->flags) == NTP_FLAG_LI_NOTINSYNC) {
+		DBG("ignoring unsynchronized peer");
 		return;
+	}
+
+	if (NTP_FLAGS_VN_DECODE(msg->flags) != 4) {
+		DBG("unsupported version %d", NTP_FLAGS_VN_DECODE(msg->flags));
+		return;
+	}
+
+	if (NTP_FLAGS_MD_DECODE(msg->flags) != NTP_FLAG_MD_SERVER) {
+		DBG("unsupported mode %d", NTP_FLAGS_MD_DECODE(msg->flags));
+		return;
+	}
 
 	org = transmit_timeval.tv_sec +
 			(1.0e-6 * transmit_timeval.tv_usec) + OFFSET_1900_1970;
@@ -182,8 +276,8 @@ static void decode_msg(void *base, size_t len, struct timeval *tv)
 	DBG("offset=%f delay=%f", offset, delay);
 
 	/* Remove the timeout, as timeserver has responded */
-	if (timeout_id > 0)
-		g_source_remove(timeout_id);
+
+	reset_timeout();
 
 	/*
 	 * Now poll the server every transmit_delay seconds
@@ -196,6 +290,8 @@ static void decode_msg(void *base, size_t len, struct timeval *tv)
 
 	poll_id = g_timeout_add_seconds(transmit_delay, next_poll, NULL);
 
+	connman_info("ntp: time slew %+.6f s", offset);
+
 	if (offset < STEPTIME_MIN_OFFSET && offset > -STEPTIME_MIN_OFFSET) {
 		struct timeval adj;
 
@@ -358,13 +454,6 @@ int __connman_ntp_start(char *server)
 
 	start_ntp(timeserver);
 
-	/*
-	 * Add a fallback timeout , preferably short, 5 sec here,
-	 * to fallback on the next server.
-	 */
-
-	timeout_id = g_timeout_add_seconds(5, next_server, NULL);
-
 	return 0;
 }
 
@@ -375,8 +464,7 @@ void __connman_ntp_stop()
 	if (poll_id > 0)
 		g_source_remove(poll_id);
 
-	if (timeout_id > 0)
-		g_source_remove(timeout_id);
+        reset_timeout();
 
 	if (channel_watch > 0) {
 		g_source_remove(channel_watch);
